#include <stdio.h>
#include <stdlib.h>
#include <elf.h>
#include <string.h>
#include <assert.h>


typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;


const char *gInputFilename = NULL;

const char *gOutputFilename = NULL;


void usage(void)
{
	printf( "elfpatch by gfoot@users.sf.net\n" );
	printf( "\n" );
	printf( "elfpatch strips all sections except for .text, makes .text writable,\n" );
	printf( "and removes section headers.  It's intended for ARM architectures\n" );
	printf( "(GP2X)	and only supports the elf file layouts I'm seeing from the \n" );
	printf( "Open2X toolchain (gcc 4 series).  YMMV with other compilers.\n" );
	printf( "\n" );
	printf( "Usage: elfpatch infile [outfile]\n" );
	printf( "\n" );
	printf( "The default is to overwrite the input file.\n" );
}

void process_args( const char *argv[] )
{
	int i;
	for( i = 1; argv[i]; ++i )
	{
		if( argv[i][0] == '-' )
		{
			switch( argv[i][1] )
			{
				case '-':
					if( !strcmp( argv[i], "--help" ) )
					{
						usage();
						exit(0);
						return;
					}
					else
					{
						goto unrecognized;
					}
					break;

				case 'h':
					usage();
					exit(0);
					return;

				default:
					goto unrecognized;
			}
		}
		else
		{
			if( !gInputFilename )
			{
				gInputFilename = argv[i];
			}
			else if( !gOutputFilename )
			{
				gOutputFilename = argv[i];
			}
			else
			{
				fprintf( stderr, "Error: too many filenames ('%s')\n", argv[i] );
				usage();
				exit(1);
				return;
			}
		}
	}

	if( !gInputFilename )
	{
		fprintf( stderr, "No input filename given\n" );
		usage();
		exit(2);
		return;
	}

	if( !gOutputFilename )
	{
		gOutputFilename = gInputFilename;
	}

	return;

unrecognized:
	fprintf( stderr, "Unrecognized option '%s'\n", argv[i] );
	usage();
	exit(3);
	return;
}


int main( int argc, const char *argv[] )
{
	process_args( argv );

	FILE *fp = fopen( gInputFilename, "rb" );
	if( !fp )
	{
		perror( gInputFilename );
		exit(4);
	}

	fseek( fp, 0, SEEK_END );
	int size = ftell( fp );

	u8 *elf = new u8[ size ];
	fseek( fp, 0, SEEK_SET );
	fread( elf, 1, size, fp );
	fclose( fp );


	// check we understand the elf header
	const Elf32_Ehdr *elf_header = (const Elf32_Ehdr *)elf;
	assert( size >= (int)sizeof *elf_header );
	assert( elf_header->e_ehsize == sizeof *elf_header );

	// find the first program header and check we understand it
	const Elf32_Phdr *prog_header = (const Elf32_Phdr *)(elf + elf_header->e_phoff);
	assert( elf_header->e_phentsize == sizeof *prog_header );
	assert( size >= (int)elf_header->e_phoff + elf_header->e_phentsize * elf_header->e_phnum );
	assert( prog_header->p_type == PT_LOAD );
	assert( prog_header->p_offset == 0 );
	assert( prog_header->p_flags & PF_X );
	assert( prog_header->p_filesz == prog_header->p_memsz );

	// skip any further program headers, to find the code
	const u8 *code = (const u8 *)(prog_header + elf_header->e_phnum );
	assert( code - elf < size );
	int code_size = prog_header->p_memsz - (code-elf);

	// get the entry point relative to the start of the code
	int code_entry = elf_header->e_entry - prog_header->p_vaddr - (code - elf);
	assert( code_entry >= 0 && code_entry < code_size );


	// construct output elf file
	int out_size;
	u8 *out_elf;
	
#if 1

	// Maybe the code is actually a raw elf image itself
	if( *(u32*)code == *(u32*)elf_header )
	{

		printf( "elfpatch: looks like you wrote your own ELF headers...\n" );
		out_size = code_size;
		out_elf = new u8[ out_size ];
		memcpy( out_elf, code, out_size );

	} else {

		printf( "elfpatch: stripping cruft\n" );

		// Construct the output elf file
		out_size = sizeof *elf_header + sizeof *prog_header + code_size;
		out_elf = new u8[ out_size ];

		// set up overlay structs
		Elf32_Ehdr *out_elf_header = (Elf32_Ehdr *)out_elf;

		Elf32_Phdr *out_prog_header = (Elf32_Phdr *)(out_elf_header + 1);

		// maybe we can overlap the prog header with the elf header?
		if( prog_header->p_type == 1 && prog_header->p_offset == 0 )
		{
			// readjust to overlap 8 bytes
			printf( "elfpatch: overlapping program header with last 8 bytes of elf header\n" );
			out_prog_header = (Elf32_Phdr *)(((u8 *)out_prog_header) - 8);
			out_size -= 8;
		}

		u8 *out_code = (u8 *)(out_prog_header + 1);

		// start by copying the input data
		memcpy( out_elf_header, elf_header, sizeof *out_elf_header );
		memcpy( out_prog_header, prog_header, sizeof *out_prog_header );
		memcpy( out_code, code, code_size );

		// touch up offsets accordingly
		out_elf_header->e_phoff = (u8 *)out_prog_header - out_elf;
		out_elf_header->e_entry = out_prog_header->p_vaddr + (out_code - out_elf) + code_entry;
		out_elf_header->e_shoff = 0;
		out_elf_header->e_phnum = 1;
		out_elf_header->e_shnum = 0;
		out_elf_header->e_shstrndx = SHN_UNDEF;

		out_prog_header->p_filesz = out_size;
		out_prog_header->p_memsz = out_size;
		out_prog_header->p_flags |= PF_W; // writable .text section

	}

#else

	// construct output elf file
	out_size = sizeof *elf_header + code_size;
	out_elf = new u8[ out_size ];

	// elf header must be at start of file
	Elf32_Ehdr *out_elf_header = (Elf32_Ehdr *)out_elf;

	// program header can overlay elf header if you're careful, as per http://www.muppetlabs.com/~breadbox/software/tiny/teensy.html
	Elf32_Phdr *out_prog_header = (Elf32_Phdr *)(out_elf+4);

	// start by copying the input elf header
	memcpy( out_elf_header, elf_header, sizeof *out_elf_header );

	out_elf_header->e_phoff = 4;

	// update fields in program header
	out_prog_header->p_type = 1;
	out_prog_header->p_offset = 0;
	out_prog_header->p_vaddr = 0x00001000;
	//out_prog_header->p_paddr = ...;
	out_prog_header->p_filesz = out_size;
	out_prog_header->

#endif

	// write to disk
	fp = fopen( gOutputFilename, "wb" );
	if( !fp )
	{
		perror( gOutputFilename );
		exit(5);
	}
	fwrite( out_elf, 1, out_size, fp );
	fclose( fp );

	return 0;
}

